#!/usr/bin/env python
# coding=utf-8

import json
from threading import Lock
import warnings

from sacred.commandline_options import cli_option
from sacred.observers.base import RunObserver
from sacred.serializer import flatten

DEFAULT_SQL_PRIORITY = 40


# ############################# Observer #################################### #


class SqlObserver(RunObserver):
    @classmethod
    def create(cls, url, echo=False, priority=DEFAULT_SQL_PRIORITY):
        warnings.warn(
            "SqlObserver.create(...) is deprecated. Please use"
            " SqlObserver(...) instead.",
            DeprecationWarning,
        )
        return cls(url, echo, priority)

    def __init__(self, url, echo=False, priority=DEFAULT_SQL_PRIORITY):
        from sqlalchemy.orm import sessionmaker, scoped_session
        import sqlalchemy as sa

        engine = sa.create_engine(url, echo=echo)
        session_factory = sessionmaker(bind=engine)
        # make session thread-local to avoid problems with sqlite (see #275)
        session = scoped_session(session_factory)
        self.engine = engine
        self.session = session
        self.priority = priority
        self.run = None
        self.lock = Lock()

    @classmethod
    def create_from(cls, engine, session, priority=DEFAULT_SQL_PRIORITY):
        """Instantiate a SqlObserver with an existing engine and session."""
        self = cls.__new__(cls)  # skip __init__ call
        self.engine = engine
        self.session = session
        self.priority = priority
        self.run = None
        self.lock = Lock()
        return self

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        return self._add_event(
            ex_info,
            command,
            host_info,
            config,
            meta_info,
            _id,
            "RUNNING",
            start_time=start_time,
        )

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        return self._add_event(
            ex_info, command, host_info, config, meta_info, _id, "QUEUED"
        )

    def _add_event(
        self, ex_info, command, host_info, config, meta_info, _id, status, **kwargs
    ):
        from .sql_bases import Base, Experiment, Host, Run

        Base.metadata.create_all(self.engine)
        sql_exp = Experiment.get_or_create(ex_info, self.session)
        sql_host = Host.get_or_create(host_info, self.session)
        if _id is None:
            i = self.session.query(Run).order_by(Run.id.desc()).first()
            _id = 0 if i is None else i.id + 1

        self.run = Run(
            run_id=str(_id),
            config=json.dumps(flatten(config)),
            command=command,
            priority=meta_info.get("priority", 0),
            comment=meta_info.get("comment", ""),
            experiment=sql_exp,
            host=sql_host,
            status=status,
            **kwargs,
        )
        self.session.add(self.run)
        self.save()
        return _id or self.run.run_id

    def heartbeat_event(self, info, captured_out, beat_time, result):
        self.run.info = json.dumps(flatten(info))
        self.run.captured_out = captured_out
        self.run.heartbeat = beat_time
        self.run.result = result
        self.save()

    def completed_event(self, stop_time, result):
        self.run.stop_time = stop_time
        self.run.result = result
        self.run.status = "COMPLETED"
        self.save()

    def interrupted_event(self, interrupt_time, status):
        self.run.stop_time = interrupt_time
        self.run.status = status
        self.save()

    def failed_event(self, fail_time, fail_trace):
        self.run.stop_time = fail_time
        self.run.fail_trace = "\n".join(fail_trace)
        self.run.status = "FAILED"
        self.save()

    def resource_event(self, filename):
        from .sql_bases import Resource

        res = Resource.get_or_create(filename, self.session)
        self.run.resources.append(res)
        self.save()

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        from .sql_bases import Artifact

        a = Artifact.create(name, filename)
        self.run.artifacts.append(a)
        self.save()

    def save(self):
        with self.lock:
            self.session.commit()

    def query(self, _id):
        from .sql_bases import Run

        run = self.session.query(Run).filter_by(id=_id).first()
        return run.to_json()

    def __eq__(self, other):
        if isinstance(other, SqlObserver):
            # fixme: this will probably fail to detect two equivalent engines
            return self.engine == other.engine and self.session == other.session
        return False


# ######################## Commandline Option ############################### #


@cli_option("-s", "--sql")
def sql_option(args, run):
    """Add a SQL Observer to the experiment.

    The typical form is: dialect://username:password@host:port/database
    """
    run.observers.append(SqlObserver(args))



#!/usr/bin/env python
# coding=utf-8

from __future__ import division, print_function, unicode_literals, absolute_import

import os
import textwrap
import uuid
from collections import OrderedDict
import warnings

from sacred.__about__ import __version__
from sacred.commandline_options import cli_option
from sacred.observers import RunObserver


class TinyDbObserver(RunObserver):

    VERSION = "TinyDbObserver-{}".format(__version__)

    @classmethod
    def create(cls, path="./runs_db", overwrite=None):
        warnings.warn(
            "TinyDbObserver.create(...) is deprecated. "
            "Please use TinyDbObserver(...) instead.",
            DeprecationWarning,
        )
        return cls(path, overwrite)

    def __init__(self, path="./runs_db", overwrite=None):
        from .tinydb_hashfs_bases import get_db_file_manager

        root_dir = os.path.abspath(path)
        os.makedirs(root_dir, exist_ok=True)

        db, fs = get_db_file_manager(root_dir)
        self.db = db
        self.runs = db.table("runs")
        self.fs = fs
        self.overwrite = overwrite
        self.run_entry = {}
        self.db_run_id = None
        self.root = root_dir

    @classmethod
    def create_from(cls, db, fs, overwrite=None, root=None):
        """Instantiate a TinyDbObserver with an existing db and filesystem."""
        self = cls.__new__(cls)  # skip __init__ call
        self.db = db
        self.runs = db.table("runs")
        self.fs = fs
        self.overwrite = overwrite
        self.run_entry = {}
        self.db_run_id = None
        self.root = root
        return self

    def save(self):
        """Insert or update the current run entry."""
        if self.db_run_id:
            self.runs.update(self.run_entry, eids=[self.db_run_id])
        else:
            db_run_id = self.runs.insert(self.run_entry)
            self.db_run_id = db_run_id

    def save_sources(self, ex_info):
        from .tinydb_hashfs_bases import BufferedReaderWrapper

        source_info = []
        for source_name, md5 in ex_info["sources"]:

            # Substitute any HOME or Environment Vars to get absolute path
            abs_path = os.path.join(ex_info["base_dir"], source_name)
            abs_path = os.path.expanduser(abs_path)
            abs_path = os.path.expandvars(abs_path)
            handle = BufferedReaderWrapper(open(abs_path, "rb"))

            file = self.fs.get(md5)
            if file:
                id_ = file.id
            else:
                address = self.fs.put(abs_path)
                id_ = address.id
            source_info.append([source_name, id_, handle])
        return source_info

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        raise NotImplementedError(
            "queued_event method is not implemented for" " local TinyDbObserver."
        )

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        self.db_run_id = None

        self.run_entry = {
            "experiment": dict(ex_info),
            "format": self.VERSION,
            "command": command,
            "host": dict(host_info),
            "start_time": start_time,
            "config": config,
            "meta": meta_info,
            "status": "RUNNING",
            "resources": [],
            "artifacts": [],
            "captured_out": "",
            "info": {},
            "heartbeat": None,
        }

        # set ID if not given
        if _id is None:
            _id = uuid.uuid4().hex

        self.run_entry["_id"] = _id

        # save sources
        self.run_entry["experiment"]["sources"] = self.save_sources(ex_info)
        self.save()
        return self.run_entry["_id"]

    def heartbeat_event(self, info, captured_out, beat_time, result):
        self.run_entry["info"] = info
        self.run_entry["captured_out"] = captured_out
        self.run_entry["heartbeat"] = beat_time
        self.run_entry["result"] = result
        self.save()

    def completed_event(self, stop_time, result):
        self.run_entry["stop_time"] = stop_time
        self.run_entry["result"] = result
        self.run_entry["status"] = "COMPLETED"
        self.save()

    def interrupted_event(self, interrupt_time, status):
        self.run_entry["stop_time"] = interrupt_time
        self.run_entry["status"] = status
        self.save()

    def failed_event(self, fail_time, fail_trace):
        self.run_entry["stop_time"] = fail_time
        self.run_entry["status"] = "FAILED"
        self.run_entry["fail_trace"] = fail_trace
        self.save()

    def resource_event(self, filename):
        from .tinydb_hashfs_bases import BufferedReaderWrapper

        id_ = self.fs.put(filename).id
        handle = BufferedReaderWrapper(open(filename, "rb"))
        resource = [filename, id_, handle]

        if resource not in self.run_entry["resources"]:
            self.run_entry["resources"].append(resource)
            self.save()

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        from .tinydb_hashfs_bases import BufferedReaderWrapper

        id_ = self.fs.put(filename).id
        handle = BufferedReaderWrapper(open(filename, "rb"))
        artifact = [name, filename, id_, handle]

        if artifact not in self.run_entry["artifacts"]:
            self.run_entry["artifacts"].append(artifact)
            self.save()

    def __eq__(self, other):
        if isinstance(other, TinyDbObserver):
            return self.runs.all() == other.runs.all()
        return False


@cli_option("-t", "--tiny_db")
def tiny_db_option(args, run):
    """Add a TinyDB Observer to the experiment.

    The argument is the path to be given to the TinyDbObserver.
    """
    tinydb_obs = TinyDbObserver(path=args)
    run.observers.append(tinydb_obs)


class TinyDbReader:
    def __init__(self, path):
        from .tinydb_hashfs_bases import get_db_file_manager

        root_dir = os.path.abspath(path)
        if not os.path.exists(root_dir):
            raise IOError("Path does not exist: %s" % path)

        db, fs = get_db_file_manager(root_dir)

        self.db = db
        self.runs = db.table("runs")
        self.fs = fs

    def search(self, *args, **kwargs):
        """Wrapper to TinyDB's search function."""
        return self.runs.search(*args, **kwargs)

    def fetch_files(self, exp_name=None, query=None, indices=None):
        """Return Dictionary of files for experiment name or query.

        Returns a list of one dictionary per matched experiment. The
        dictionary is of the following structure

            {
              'exp_name': 'scascasc',
              'exp_id': 'dqwdqdqwf',
              'date': datatime_object,
              'sources': [ {'filename': filehandle}, ..., ],
              'resources': [ {'filename': filehandle}, ..., ],
              'artifacts': [ {'filename': filehandle}, ..., ]
            }

        """
        entries = self.fetch_metadata(exp_name, query, indices)

        all_matched_entries = []
        for ent in entries:

            rec = dict(
                exp_name=ent["experiment"]["name"],
                exp_id=ent["_id"],
                date=ent["start_time"],
            )

            source_files = {x[0]: x[2] for x in ent["experiment"]["sources"]}
            resource_files = {x[0]: x[2] for x in ent["resources"]}
            artifact_files = {x[0]: x[3] for x in ent["artifacts"]}

            if source_files:
                rec["sources"] = source_files
            if resource_files:
                rec["resources"] = resource_files
            if artifact_files:
                rec["artifacts"] = artifact_files

            all_matched_entries.append(rec)

        return all_matched_entries

    def fetch_report(self, exp_name=None, query=None, indices=None):

        template = """
-------------------------------------------------
Experiment: {exp_name}
-------------------------------------------------
ID: {exp_id}
Date: {start_date}    Duration: {duration}

Parameters:
{parameters}

Result:
{result}

Dependencies:
{dependencies}

Resources:
{resources}

Source Files:
{sources}

Outputs:
{artifacts}
"""

        entries = self.fetch_metadata(exp_name, query, indices)

        all_matched_entries = []
        for ent in entries:

            date = ent["start_time"]
            weekdays = "Mon Tue Wed Thu Fri Sat Sun".split()
            w = weekdays[date.weekday()]
            date = " ".join([w, date.strftime("%d %b %Y")])

            duration = ent["stop_time"] - ent["start_time"]
            secs = duration.total_seconds()
            hours, remainder = divmod(secs, 3600)
            minutes, seconds = divmod(remainder, 60)
            duration = "%02d:%02d:%04.1f" % (hours, minutes, seconds)

            parameters = self._dict_to_indented_list(ent["config"])

            result = self._indent(ent["result"].__repr__(), prefix="    ")

            deps = ent["experiment"]["dependencies"]
            deps = self._indent("\n".join(deps), prefix="    ")

            resources = [x[0] for x in ent["resources"]]
            resources = self._indent("\n".join(resources), prefix="    ")

            sources = [x[0] for x in ent["experiment"]["sources"]]
            sources = self._indent("\n".join(sources), prefix="    ")

            artifacts = [x[0] for x in ent["artifacts"]]
            artifacts = self._indent("\n".join(artifacts), prefix="    ")

            none_str = "    None"

            rec = dict(
                exp_name=ent["experiment"]["name"],
                exp_id=ent["_id"],
                start_date=date,
                duration=duration,
                parameters=parameters if parameters else none_str,
                result=result if result else none_str,
                dependencies=deps if deps else none_str,
                resources=resources if resources else none_str,
                sources=sources if sources else none_str,
                artifacts=artifacts if artifacts else none_str,
            )

            report = template.format(**rec)

            all_matched_entries.append(report)

        return all_matched_entries

    def fetch_metadata(self, exp_name=None, query=None, indices=None):
        """Return all metadata for matching experiment name, index or query."""
        from tinydb import Query
        from tinydb.queries import QueryImpl

        if exp_name or query:
            if query:
                assert type(query), QueryImpl
                q = query
            elif exp_name:
                q = Query().experiment.name.search(exp_name)

            entries = self.runs.search(q)

        elif indices or indices == 0:
            if not isinstance(indices, (tuple, list)):
                indices = [indices]

            num_recs = len(self.runs)

            for idx in indices:
                if idx >= num_recs:
                    raise ValueError(
                        "Index value ({}) must be less than "
                        "number of records ({})".format(idx, num_recs)
                    )

            entries = [self.runs.all()[ind] for ind in indices]

        else:
            raise ValueError(
                "Must specify an experiment name, indicies or " "pass custom query"
            )

        return entries

    def _dict_to_indented_list(self, d):

        d = OrderedDict(sorted(d.items(), key=lambda t: t[0]))

        output_str = ""

        for k, v in d.items():
            output_str += "%s: %s" % (k, v)
            output_str += "\n"

        output_str = self._indent(output_str.strip(), prefix="    ")

        return output_str

    def _indent(self, message, prefix):
        """Wrapper for indenting strings in Python 2 and 3."""
        preferred_width = 150
        wrapper = textwrap.TextWrapper(
            initial_indent=prefix, width=preferred_width, subsequent_indent=prefix
        )

        lines = message.splitlines()
        formatted_lines = [wrapper.fill(lin) for lin in lines]
        formatted_text = "\n".join(formatted_lines)

        return formatted_text



#!/usr/bin/env python
# coding=utf-8

import json
import os
import os.path
from pathlib import Path
from typing import Optional
import warnings

from shutil import copyfile

from sacred.commandline_options import cli_option
from sacred.dependencies import get_digest
from sacred.observers.base import RunObserver
from sacred import optional as opt
from sacred.serializer import flatten
from sacred.utils import PathType


DEFAULT_FILE_STORAGE_PRIORITY = 20


class FileStorageObserver(RunObserver):
    VERSION = "FileStorageObserver-0.7.0"

    @classmethod
    def create(cls, *args, **kwargs):
        warnings.warn(
            "FileStorageObserver.create(...) is deprecated. "
            "Please use FileStorageObserver(...) instead.",
            DeprecationWarning,
        )
        return cls(*args, **kwargs)

    def __init__(
        self,
        basedir: PathType,
        resource_dir: Optional[PathType] = None,
        source_dir: Optional[PathType] = None,
        template: Optional[PathType] = None,
        priority: int = DEFAULT_FILE_STORAGE_PRIORITY,
    ):
        basedir = Path(basedir)
        resource_dir = resource_dir or basedir / "_resources"
        source_dir = source_dir or basedir / "_sources"
        if template is not None:
            if not os.path.exists(template):
                raise FileNotFoundError(
                    "Couldn't find template file '{}'".format(template)
                )
        else:
            template = basedir / "template.html"
            if not template.exists():
                template = None
        self.initialize(basedir, resource_dir, source_dir, template, priority)

    def initialize(
        self,
        basedir,
        resource_dir,
        source_dir,
        template,
        priority=DEFAULT_FILE_STORAGE_PRIORITY,
    ):
        self.basedir = str(basedir)
        self.resource_dir = resource_dir
        self.source_dir = source_dir
        self.template = template
        self.priority = priority
        self.dir = None
        self.run_entry = None
        self.config = None
        self.info = None
        self.cout = ""
        self.cout_write_cursor = 0

    @classmethod
    def create_from(cls, *args, **kwargs):
        self = cls.__new__(cls)  # skip __init__ call
        self.initialize(*args, **kwargs)
        return self

    def _maximum_existing_run_id(self):
        dir_nrs = [
            int(d)
            for d in os.listdir(self.basedir)
            if os.path.isdir(os.path.join(self.basedir, d)) and d.isdigit()
        ]
        if dir_nrs:
            return max(dir_nrs)
        else:
            return 0

    def _make_dir(self, _id):
        new_dir = os.path.join(self.basedir, str(_id))
        os.mkdir(new_dir)
        self.dir = new_dir  # set only if mkdir is successful

    def _make_run_dir(self, _id):
        os.makedirs(self.basedir, exist_ok=True)
        self.dir = None
        if _id is None:
            fail_count = 0
            _id = self._maximum_existing_run_id() + 1
            while self.dir is None:
                try:
                    self._make_dir(_id)
                except FileExistsError:  # Catch race conditions
                    if fail_count < 1000:
                        fail_count += 1
                        _id += 1
                    else:  # expect that something else went wrong
                        raise
        else:
            self.dir = os.path.join(self.basedir, str(_id))
            os.mkdir(self.dir)

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        self._make_run_dir(_id)

        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "meta": meta_info,
            "status": "QUEUED",
        }
        self.config = config
        self.info = {}

        self.save_json(self.run_entry, "run.json")
        self.save_json(self.config, "config.json")

        for s, m in ex_info["sources"]:
            self.save_file(s)

        return os.path.relpath(self.dir, self.basedir) if _id is None else _id

    def save_sources(self, ex_info):
        base_dir = ex_info["base_dir"]
        source_info = []
        for s, m in ex_info["sources"]:
            abspath = os.path.join(base_dir, s)
            store_path, md5sum = self.find_or_save(abspath, self.source_dir)
            # assert m == md5sum
            relative_source = os.path.relpath(str(store_path), self.basedir)
            source_info.append([s, relative_source])
        return source_info

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        self._make_run_dir(_id)

        ex_info["sources"] = self.save_sources(ex_info)

        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "start_time": start_time.isoformat(),
            "meta": meta_info,
            "status": "RUNNING",
            "resources": [],
            "artifacts": [],
            "heartbeat": None,
        }
        self.config = config
        self.info = {}
        self.cout = ""
        self.cout_write_cursor = 0

        self.save_json(self.run_entry, "run.json")
        self.save_json(self.config, "config.json")
        self.save_cout()

        return os.path.relpath(self.dir, self.basedir) if _id is None else _id

    def find_or_save(self, filename, store_dir: Path):
        os.makedirs(str(store_dir), exist_ok=True)
        source_name, ext = os.path.splitext(os.path.basename(filename))
        md5sum = get_digest(filename)
        store_name = source_name + "_" + md5sum + ext
        store_path = store_dir / store_name
        if not store_path.exists():
            copyfile(filename, str(store_path))
        return store_path, md5sum

    def save_json(self, obj, filename):
        with open(os.path.join(self.dir, filename), "w") as f:
            json.dump(flatten(obj), f, sort_keys=True, indent=2)

    def save_file(self, filename, target_name=None):
        target_name = target_name or os.path.basename(filename)
        blacklist = ["run.json", "config.json", "cout.txt", "metrics.json"]
        blacklist = [os.path.join(self.dir, x) for x in blacklist]
        dest_file = os.path.join(self.dir, target_name)
        if dest_file in blacklist:
            raise FileExistsError(
                "You are trying to overwrite a file necessary for the "
                "FileStorageObserver. "
                "The list of blacklisted files is: {}".format(blacklist)
            )
        copyfile(filename, dest_file)

    def save_cout(self):
        with open(os.path.join(self.dir, "cout.txt"), "ab") as f:
            f.write(self.cout[self.cout_write_cursor :].encode("utf-8"))
            self.cout_write_cursor = len(self.cout)

    def render_template(self):
        if opt.has_mako and self.template:
            from mako.template import Template

            template = Template(filename=self.template)
            report = template.render(
                run=self.run_entry,
                config=self.config,
                info=self.info,
                cout=self.cout,
                savedir=self.dir,
            )
            ext = self.template.suffix
            with open(os.path.join(self.dir, "report" + ext), "w") as f:
                f.write(report)

    def heartbeat_event(self, info, captured_out, beat_time, result):
        self.info = info
        self.run_entry["heartbeat"] = beat_time.isoformat()
        self.run_entry["result"] = result
        self.cout = captured_out
        self.save_cout()
        self.save_json(self.run_entry, "run.json")
        if self.info:
            self.save_json(self.info, "info.json")

    def completed_event(self, stop_time, result):
        self.run_entry["stop_time"] = stop_time.isoformat()
        self.run_entry["result"] = result
        self.run_entry["status"] = "COMPLETED"

        self.save_json(self.run_entry, "run.json")
        self.render_template()

    def interrupted_event(self, interrupt_time, status):
        self.run_entry["stop_time"] = interrupt_time.isoformat()
        self.run_entry["status"] = status
        self.save_json(self.run_entry, "run.json")
        self.render_template()

    def failed_event(self, fail_time, fail_trace):
        self.run_entry["stop_time"] = fail_time.isoformat()
        self.run_entry["status"] = "FAILED"
        self.run_entry["fail_trace"] = fail_trace
        self.save_json(self.run_entry, "run.json")
        self.render_template()

    def resource_event(self, filename):
        store_path, md5sum = self.find_or_save(filename, self.resource_dir)
        self.run_entry["resources"].append([filename, str(store_path)])
        self.save_json(self.run_entry, "run.json")

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        self.save_file(filename, name)
        self.run_entry["artifacts"].append(name)
        self.save_json(self.run_entry, "run.json")

    def log_metrics(self, metrics_by_name, info):
        """Store new measurements into metrics.json."""
        try:
            metrics_path = os.path.join(self.dir, "metrics.json")
            with open(metrics_path, "r") as f:
                saved_metrics = json.load(f)
        except IOError:
            # We haven't recorded anything yet. Start Collecting.
            saved_metrics = {}

        for metric_name, metric_ptr in metrics_by_name.items():

            if metric_name not in saved_metrics:
                saved_metrics[metric_name] = {
                    "values": [],
                    "steps": [],
                    "timestamps": [],
                }

            saved_metrics[metric_name]["values"] += metric_ptr["values"]
            saved_metrics[metric_name]["steps"] += metric_ptr["steps"]

            # Manually convert them to avoid passing a datetime dtype handler
            # when we're trying to convert into json.
            timestamps_norm = [ts.isoformat() for ts in metric_ptr["timestamps"]]
            saved_metrics[metric_name]["timestamps"] += timestamps_norm

        self.save_json(saved_metrics, "metrics.json")

    def __eq__(self, other):
        if isinstance(other, FileStorageObserver):
            return self.basedir == other.basedir
        return False


@cli_option("-F", "--file_storage")
def file_storage_option(args, run):
    """Add a file-storage observer to the experiment.

    The value of the arguement should be the
    base-directory to write the runs to
    """
    run.observers.append(FileStorageObserver(args))



#!/usr/bin/env python
# coding=utf-8

import importlib
from sacred.utils import modules_exist
from sacred.utils import get_package_version, parse_version


def optional_import(*package_names):
    try:
        packages = [importlib.import_module(pn) for pn in package_names]
        return True, packages[0]
    except ImportError:
        return False, None


def get_tensorflow():
    # Ensures backward and forward compatibility with TensorFlow 1 and 2.
    if get_package_version("tensorflow") < parse_version("1.13.1"):
        import warnings

        warnings.warn(
            "Use of TensorFlow 1.12 and older is deprecated. "
            "Use Tensorflow 1.13 or newer instead.",
            DeprecationWarning,
        )
        import tensorflow as tf
    else:
        import tensorflow.compat.v1 as tf
    return tf


# Get libc in a cross-platform way and use it to also flush the c stdio buffers
# credit to J.F. Sebastians SO answer from here:
# http://stackoverflow.com/a/22434262/1388435
try:
    import ctypes
    from ctypes.util import find_library
except ImportError:
    libc = None
else:
    try:
        libc = ctypes.cdll.msvcrt  # Windows
    except OSError:
        libc = ctypes.cdll.LoadLibrary(find_library("c"))


has_numpy, np = optional_import("numpy")
has_yaml, yaml = optional_import("yaml")
has_pandas, pandas = optional_import("pandas")

has_sqlalchemy = modules_exist("sqlalchemy")
has_mako = modules_exist("mako")
has_tinydb = modules_exist("tinydb", "tinydb_serialization", "hashfs")
has_tensorflow = modules_exist("tensorflow")



"""
Provides the basis for all command-line options (flags) in sacred.

It defines the base class CommandLineOption and the standard supported flags.
Some further options that add observers to the run are defined alongside those.
"""

from typing import Callable
import inspect
import re

from sacred.run import Run
from sacred.commands import print_config
from sacred.utils import convert_camel_case_to_snake_case


CLIFunction = Callable[[str, Run], None]


class CLIOption:
    def __init__(
        self,
        apply_function: CLIFunction,
        short_flag: str,
        long_flag: str,
        is_flag: bool,
    ):

        if not re.match(r"-\w$", short_flag):
            raise ValueError(
                "Short flag malformed. " "One correct short flag would be: `-j`."
            )
        if not re.match(r"--[\w-]+\w$", long_flag):
            raise ValueError(
                "Long flag malformed. One correct long flag "
                "would be: `--my-pretty-flag`"
            )
        self.apply_function = apply_function
        self.short_flag = short_flag
        self.long_flag = long_flag
        self.is_flag = is_flag

        # trick for backward compatibility:
        self.arg = None if is_flag else "VALUE"
        self.arg_description = None if is_flag else ""

    def __call__(self, *args, **kwargs):
        return self.apply_function(*args, **kwargs)

    def get_flag(self):
        """Legacy function. Should be removed at some point."""
        return self.long_flag

    def get_short_flag(self):
        """Legacy function. Should be removed at some point."""
        return self.short_flag

    def get_flags(self):
        """Legacy function. Should be removed at some point."""
        return self.short_flag, self.long_flag

    def apply(self, args, run):
        """Legacy function. Should be removed at some point."""
        return self.apply_function(args, run)

    def get_name(self):
        return self.apply_function.__name__

    def get_description(self):
        return inspect.getdoc(self.apply_function) or ""


def cli_option(short_flag: str, long_flag: str, is_flag=False):
    def wrapper(f: CLIFunction):
        return CLIOption(f, short_flag, long_flag, is_flag)

    return wrapper


class CommandLineOption:
    """
    Base class for all command-line options.

    To implement a new command-line option just inherit from this class.
    Then add the `flag` class-attribute to specify the name and a class
    docstring with the description.
    If your command-line option should take an argument you must also provide
    its name via the `arg` class attribute and its description as
    `arg_description`.
    Finally you need to implement the `execute` classmethod. It receives the
    value of the argument (if applicable) and the current run. You can modify
    the run object in any way.

    If the command line option depends on one or more installed packages, those
    should be imported in the `apply` method to get a proper ImportError
    if the packages are not available.
    """

    _enabled = True

    short_flag = None
    """ The (one-letter) short form (defaults to first letter of flag) """

    arg = None
    """ Name of the argument (optional) """

    arg_description = None
    """ Description of the argument (optional) """

    @classmethod
    def get_flag(cls):
        # Get the flag name from the class name
        flag = cls.__name__
        if flag.endswith("Option"):
            flag = flag[:-6]
        return "--" + convert_camel_case_to_snake_case(flag)

    @classmethod
    def get_short_flag(cls):
        if cls.short_flag is None:
            return "-" + cls.get_flag()[2]
        else:
            return "-" + cls.short_flag

    @classmethod
    def get_flags(cls):
        """
        Return the short and the long version of this option.

        The long flag (e.g. '--foo_bar'; used on the command-line like this:
        --foo_bar[=ARGS]) is derived from the class-name by stripping away any
        -Option suffix and converting the rest to snake_case.

        The short flag (e.g. '-f'; used on the command-line like this:
        -f [ARGS]) the short_flag class-member if that is set, or the first
        letter of the long flag otherwise.

        Returns
        -------
        (str, str)
            tuple of short-flag, and long-flag

        """
        return cls.get_short_flag(), cls.get_flag()

    @classmethod
    def apply(cls, args, run):
        """
        Modify the current Run base on this command-line option.

        This function is executed after constructing the Run object, but
        before actually starting it.

        Parameters
        ----------
        args : bool | str
            If this command-line option accepts an argument this will be value
            of that argument if set or None.
            Otherwise it is either True or False.
        run :  sacred.run.Run
            The current run to be modified

        """
        pass


def get_name(option):
    if isinstance(option, CLIOption):
        return option.get_name()
    else:
        return option.__name__


@cli_option("-h", "--help", is_flag=True)
def help_option(args, run):
    """Print this help message and exit."""
    pass


@cli_option("-d", "--debug", is_flag=True)
def debug_option(args, run):
    """
    Set this run to debug mode.

    Suppress warnings about missing observers and don't filter the stacktrace.
    Also enables usage with ipython `--pdb`.
    """
    run.debug = True


@cli_option("-D", "--pdb", is_flag=True)
def pdb_option(args, run):
    """Automatically enter post-mortem debugging with pdb on failure."""
    run.pdb = True


@cli_option("-l", "--loglevel")
def loglevel_option(args, run):
    """
    Set the LogLevel.

    Loglevel either as 0 - 50 or as string: DEBUG(10),
    INFO(20), WARNING(30), ERROR(40), CRITICAL(50)
    """
    # TODO: sacred.initialize.create_run already takes care of this

    try:
        lvl = int(args)
    except ValueError:
        lvl = args
    run.root_logger.setLevel(lvl)


@cli_option("-c", "--comment")
def comment_option(args, run):
    """Add a comment to this run."""
    run.meta_info["comment"] = args


@cli_option("-b", "--beat-interval")
def beat_interval_option(args, run):
    """
    Set the heart-beat interval for this run.

    Time between two heartbeat events is measured in seconds.
    """
    run.beat_interval = float(args)


@cli_option("-u", "--unobserved", is_flag=True)
def unobserved_option(args, run):
    """Ignore all observers for this run."""
    run.unobserved = True


@cli_option("-q", "--queue", is_flag=True)
def queue_option(args, run):
    """Only queue this run, do not start it."""
    run.queue_only = True


@cli_option("-f", "--force", is_flag=True)
def force_option(args, run):
    """Disable warnings about suspicious changes for this run."""
    run.force = True


@cli_option("-P", "--priority")
def priority_option(args, run):
    """Sets the priority for a queued up experiment.

    `--priority=NUMBER`
    The number represent the priority for this run.
    """
    try:
        priority = float(args)
    except ValueError:
        raise ValueError(
            "The PRIORITY argument must be a number! (but was '{}')".format(args)
        )
    run.meta_info["priority"] = priority


@cli_option("-e", "--enforce_clean", is_flag=True)
def enforce_clean_option(args, run):
    """Fail if any version control repository is dirty."""
    repos = run.experiment_info["repositories"]
    if not repos:
        raise RuntimeError(
            "No version control detected. "
            "Cannot enforce clean repository.\n"
            "Make sure that your sources under VCS and the "
            "corresponding python package is installed."
        )
    else:
        for repo in repos:
            if repo["dirty"]:
                raise RuntimeError(
                    "EnforceClean: Uncommited changes in "
                    'the "{}" repository.'.format(repo)
                )


@cli_option("-p", "--print-config", is_flag=True)
def print_config_option(args, run):
    """Always print the configuration first."""
    print_config(run)
    print("-" * 79)


@cli_option("-n", "--name")
def name_option(args, run):
    """Set the name for this run."""
    run.experiment_info["name"] = args
    run.run_logger = run.root_logger.getChild(args)


@cli_option("-C", "--capture")
def capture_option(args, run):
    """
    Control the way stdout and stderr are captured.

    The argument value must be one of [no, sys, fd]
    """
    run.capture_mode = args



import jsonpickle
import json as _json
from sacred import optional as opt

json = jsonpickle


__all__ = ("flatten", "restore")


if opt.has_numpy:
    import jsonpickle.ext.numpy as jsonpickle_numpy

    np = opt.np

    jsonpickle_numpy.register_handlers()

if opt.has_pandas:
    import jsonpickle.ext.pandas as jsonpickle_pandas

    jsonpickle_pandas.register_handlers()


jsonpickle.set_encoder_options("simplejson", sort_keys=True, indent=4)
jsonpickle.set_encoder_options("demjson", compactly=False)


def flatten(obj):
    return _json.loads(json.encode(obj, keys=True))


def restore(flat):
    return json.decode(_json.dumps(flat), keys=True)



#!/usr/bin/env python
# coding=utf-8

from sacred.observers.base import RunObserver, td_format
from sacred.config.config_files import load_config_file
import json


DEFAULT_SLACK_PRIORITY = 10


class SlackObserver(RunObserver):
    """Sends a message to Slack upon completion/failing of an experiment."""

    @classmethod
    def from_config(cls, filename):
        """
        Create a SlackObserver from a given configuration file.

        The file can be in any format supported by Sacred
        (.json, .pickle, [.yaml]).
        It has to specify a ``webhook_url`` and can optionally set
        ``bot_name``, ``icon``, ``completed_text``, ``interrupted_text``, and
        ``failed_text``.
        """
        return cls(**load_config_file(filename))

    def __init__(
        self,
        webhook_url,
        bot_name="sacred-bot",
        icon=":angel:",
        priority=DEFAULT_SLACK_PRIORITY,
        completed_text=None,
        interrupted_text=None,
        failed_text=None,
    ):
        self.webhook_url = webhook_url
        self.bot_name = bot_name
        self.icon = icon
        self.completed_text = completed_text or (
            ":white_check_mark: *{experiment[name]}* "
            "completed after _{elapsed_time}_ with result=`{result}`"
        )
        self.interrupted_text = interrupted_text or (
            ":warning: *{experiment[name]}* " "interrupted after _{elapsed_time}_"
        )
        self.failed_text = failed_text or (
            ":x: *{experiment[name]}* failed after " "_{elapsed_time}_ with `{error}`"
        )
        self.run = None
        self.priority = priority

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        self.run = {
            "_id": _id,
            "config": config,
            "start_time": start_time,
            "experiment": ex_info,
            "command": command,
            "host_info": host_info,
        }

    def get_completed_text(self):
        return self.completed_text.format(**self.run)

    def get_interrupted_text(self):
        return self.interrupted_text.format(**self.run)

    def get_failed_text(self):
        return self.failed_text.format(**self.run)

    def completed_event(self, stop_time, result):
        import requests

        if self.completed_text is None:
            return

        self.run["result"] = result
        self.run["stop_time"] = stop_time
        self.run["elapsed_time"] = td_format(stop_time - self.run["start_time"])

        data = {
            "username": self.bot_name,
            "icon_emoji": self.icon,
            "text": self.get_completed_text(),
        }
        headers = {"Content-type": "application/json", "Accept": "text/plain"}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)

    def interrupted_event(self, interrupt_time, status):
        import requests

        if self.interrupted_text is None:
            return

        self.run["status"] = status
        self.run["interrupt_time"] = interrupt_time
        self.run["elapsed_time"] = td_format(interrupt_time - self.run["start_time"])

        data = {
            "username": self.bot_name,
            "icon_emoji": self.icon,
            "text": self.get_interrupted_text(),
        }
        headers = {"Content-type": "application/json", "Accept": "text/plain"}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)

    def failed_event(self, fail_time, fail_trace):
        import requests

        if self.failed_text is None:
            return

        self.run["fail_trace"] = fail_trace
        self.run["error"] = fail_trace[-1].strip()
        self.run["fail_time"] = fail_time
        self.run["elapsed_time"] = td_format(fail_time - self.run["start_time"])

        data = {
            "username": self.bot_name,
            "icon_emoji": self.icon,
            "text": self.get_failed_text(),
        }
        headers = {"Content-type": "application/json", "Accept": "text/plain"}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)



from collections import namedtuple
from queue import Queue
from sacred.observers.base import RunObserver
from sacred.utils import IntervalTimer
import traceback
import logging

logger = logging.getLogger(__name__)

WrappedEvent = namedtuple("WrappedEvent", "name args kwargs")


class QueueObserver(RunObserver):
    """Wraps any observer and puts processing of events in the background.

    If the covered observer fails to process an event, the queue observer
    will retry until it works. This is useful for observers that rely on
    external services like databases that might become temporarily
    unavailable.
    """

    def __init__(
        self,
        covered_observer: RunObserver,
        interval: float = 20.0,
        retry_interval: float = 10.0,
    ):
        """Initialize QueueObserver.

        Parameters
        ----------
        covered_observer
            The real observer that is being wrapped.
        interval
            The interval in seconds at which the background thread is woken up to process new events.
        retry_interval
            The interval in seconds to wait if an event failed to be processed.
        """
        self._covered_observer = covered_observer
        self._retry_interval = retry_interval
        self._interval = interval
        self._queue = None
        self._worker = None
        self._stop_worker_event = None
        logger.debug("just testing")

    def queued_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("queued_event", args, kwargs))

    def started_event(self, *args, **kwargs):
        self._queue = Queue()
        self._stop_worker_event, self._worker = IntervalTimer.create(
            self._run, interval=self._interval
        )
        self._worker.start()

        # Putting the started event on the queue makes no sense
        # as it is required for initialization of the covered observer.
        return self._covered_observer.started_event(*args, **kwargs)

    def heartbeat_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("heartbeat_event", args, kwargs))

    def completed_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("completed_event", args, kwargs))
        self.join()

    def interrupted_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("interrupted_event", args, kwargs))
        self.join()

    def failed_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("failed_event", args, kwargs))
        self.join()

    def resource_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("resource_event", args, kwargs))

    def artifact_event(self, *args, **kwargs):
        self._queue.put(WrappedEvent("artifact_event", args, kwargs))

    def log_metrics(self, metrics_by_name, info):
        for metric_name, metric_values in metrics_by_name.items():
            self._queue.put(
                WrappedEvent("log_metrics", [metric_name, metric_values, info], {})
            )

    def _run(self):
        """Empty the queue every interval."""
        while not self._queue.empty():
            try:
                event = self._queue.get()
            except IndexError:
                # Currently there is no event on the queue so
                # just go back to sleep.
                pass
            else:
                try:
                    method = getattr(self._covered_observer, event.name)
                except NameError:
                    # The covered observer does not implement an event handler
                    # for the event, so just discard the message.
                    self._queue.task_done()
                else:
                    while True:
                        try:
                            method(*event.args, **event.kwargs)
                        except:
                            # Something went wrong during the processing of
                            # the event so wait for some time and
                            # then try again.
                            logger.debug(
                                "Error while processing event. Trying again.\n{}".format(
                                    traceback.format_exc()
                                )
                            )
                            # logging.debug(f"""Error while processing event. Trying again.
                            # {traceback.format_exc()}""")

                            self._stop_worker_event.wait(self._retry_interval)
                            continue
                        else:
                            self._queue.task_done()
                            break

    def join(self):
        if self._queue is not None:
            self._queue.join()
            self._stop_worker_event.set()
            self._worker.join(timeout=10)

    def __getattr__(self, item):
        return getattr(self._covered_observer, item)

    def __eq__(self, other):
        return self._covered_observer == other



"""Helps to collect information about the host of an experiment."""

import os
import platform
import re
import subprocess
from xml.etree import ElementTree
import warnings
from typing import List

import cpuinfo

from sacred.utils import optional_kwargs_decorator
from sacred.settings import SETTINGS

__all__ = ("host_info_gatherers", "get_host_info", "host_info_getter")

# Legacy global dict of functions that are used
# to collect the host information.
host_info_gatherers = {}


class IgnoreHostInfo(Exception):
    """Used by host_info_getters to signal that this cannot be gathered."""


class HostInfoGetter:
    def __init__(self, getter_function, name):
        self.getter_function = getter_function
        self.name = name

    def __call__(self):
        return self.getter_function()

    def get_info(self):
        return self.getter_function()


def host_info_gatherer(name):
    def wrapper(f):
        return HostInfoGetter(f, name)

    return wrapper


def check_additional_host_info(additional_host_info: List[HostInfoGetter]):
    names_taken = [x.name for x in _host_info_gatherers_list]
    for getter in additional_host_info:
        if getter.name in names_taken:
            error_msg = (
                "Key {} used in `additional_host_info` already exists as a "
                "default gatherer function. Do not use the following keys: "
                "{}"
            ).format(getter.name, names_taken)
            raise KeyError(error_msg)


def get_host_info(additional_host_info: List[HostInfoGetter] = None):
    """Collect some information about the machine this experiment runs on.

    Returns
    -------
    dict
        A dictionary with information about the CPU, the OS and the
        Python version of this machine.

    """
    additional_host_info = additional_host_info or []
    # can't use += because we don't want to modify the mutable argument.
    additional_host_info = additional_host_info + _host_info_gatherers_list
    all_host_info_gatherers = host_info_gatherers.copy()
    for getter in additional_host_info:
        all_host_info_gatherers[getter.name] = getter
    host_info = {}
    for k, v in all_host_info_gatherers.items():
        try:
            host_info[k] = v()
        except IgnoreHostInfo:
            pass
    return host_info


@optional_kwargs_decorator
def host_info_getter(func, name=None):
    """
    The decorated function is added to the process of collecting the host_info.

    This just adds the decorated function to the global
    ``sacred.host_info.host_info_gatherers`` dictionary.
    The functions from that dictionary are used when collecting the host info
    using :py:func:`~sacred.host_info.get_host_info`.

    Parameters
    ----------
    func : callable
        A function that can be called without arguments and returns some
        json-serializable information.
    name : str, optional
        The name of the corresponding entry in host_info.
        Defaults to the name of the function.

    Returns
    -------
    The function itself.

    """
    warnings.warn(
        "The host_info_getter is deprecated. "
        "Please use the `additional_host_info` argument"
        " in the Experiment constructor.",
        DeprecationWarning,
    )
    name = name or func.__name__
    host_info_gatherers[name] = func
    return func


# #################### Default Host Information ###############################


@host_info_gatherer(name="hostname")
def _hostname():
    return platform.node()


@host_info_gatherer(name="os")
def _os():
    return [platform.system(), platform.platform()]


@host_info_gatherer(name="python_version")
def _python_version():
    return platform.python_version()


@host_info_gatherer(name="cpu")
def _cpu():
    if platform.system() == "Windows":
        return _get_cpu_by_pycpuinfo()
    try:
        if platform.system() == "Darwin":
            return _get_cpu_by_sysctl()
        elif platform.system() == "Linux":
            return _get_cpu_by_proc_cpuinfo()
    except Exception:
        # Use pycpuinfo only if other ways fail, since it takes about 1 sec
        return _get_cpu_by_pycpuinfo()


@host_info_gatherer(name="gpus")
def _gpus():
    if not SETTINGS.HOST_INFO.INCLUDE_GPU_INFO:
        return

    try:
        xml = subprocess.check_output(["nvidia-smi", "-q", "-x"]).decode()
    except (FileNotFoundError, OSError, subprocess.CalledProcessError):
        raise IgnoreHostInfo()

    gpu_info = {"gpus": []}
    for child in ElementTree.fromstring(xml):
        if child.tag == "driver_version":
            gpu_info["driver_version"] = child.text
        if child.tag != "gpu":
            continue
        gpu = {
            "model": child.find("product_name").text,
            "total_memory": int(
                child.find("fb_memory_usage").find("total").text.split()[0]
            ),
            "persistence_mode": (child.find("persistence_mode").text == "Enabled"),
        }
        gpu_info["gpus"].append(gpu)

    return gpu_info


@host_info_gatherer(name="ENV")
def _environment():
    keys_to_capture = SETTINGS.HOST_INFO.CAPTURED_ENV
    return {k: os.environ[k] for k in keys_to_capture if k in os.environ}


_host_info_gatherers_list = [_hostname, _os, _python_version, _cpu, _gpus, _environment]

# ################### Get CPU Information ###############################


def _get_cpu_by_sysctl():
    os.environ["PATH"] += ":/usr/sbin"
    command = ["sysctl", "-n", "machdep.cpu.brand_string"]
    return subprocess.check_output(command).decode().strip()


def _get_cpu_by_proc_cpuinfo():
    command = ["cat", "/proc/cpuinfo"]
    all_info = subprocess.check_output(command).decode()
    model_pattern = re.compile(r"^\s*model name\s*:")
    for line in all_info.split("\n"):
        if model_pattern.match(line):
            return model_pattern.sub("", line, 1).strip()


def _get_cpu_by_pycpuinfo():
    return cpuinfo.get_cpu_info().get("brand", "Unknown")



import json
import os
import os.path
import re
from typing import Optional

from sacred.commandline_options import cli_option
from sacred.dependencies import get_digest
from sacred.observers.base import RunObserver
from sacred.serializer import flatten
from sacred.utils import PathType

DEFAULT_GCS_PRIORITY = 20


def _is_valid_bucket(bucket_name: str):
    """Validates correctness of bucket naming.

    Reference: https://cloud.google.com/storage/docs/naming
    """
    if bucket_name.startswith("gs://"):
        return False

    if len(bucket_name) < 3 or len(bucket_name) > 63:
        return False

    # IP address
    if re.match(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$", bucket_name):
        return False

    if not re.fullmatch(r"([^A-Z]|-|_|[.]|)+", bucket_name):
        return False

    if ".." in bucket_name:
        return False

    if "goog" in bucket_name or "g00g" in bucket_name:
        return False

    return True


def gcs_join(*args):
    return "/".join(args)


class GoogleCloudStorageObserver(RunObserver):
    VERSION = "GoogleCloudStorageObserver-0.1.0"

    def __init__(
        self,
        bucket: str,
        basedir: PathType,
        resource_dir: Optional[PathType] = None,
        source_dir: Optional[PathType] = None,
        priority: Optional[int] = DEFAULT_GCS_PRIORITY,
    ):
        """Constructor for a GoogleCloudStorageObserver object.

        Run when the object is first created,
        before it's used within an experiment.

        Parameters
        ----------
        bucket
            The name of the bucket you want to store results in.
            Needs to be a valid bucket name without 'gs://'
        basedir
            The relative path inside your bucket where you want this experiment to store results
        resource_dir
            Where to store resources for this experiment. By
            default, will be <basedir>/_resources
        source_dir
            Where to store code sources for this experiment. By
            default, will be <basedir>/sources
        priority
            The priority to assign to this observer if
            multiple observers are present
        """
        if not _is_valid_bucket(bucket):
            raise ValueError(
                "Your chosen bucket name doesn't follow Google Cloud Storage bucket naming rules"
            )
        resource_dir = resource_dir or "/".join([basedir, "_resources"])
        source_dir = source_dir or "/".join([basedir, "_sources"])

        self.basedir = basedir
        self.bucket_id = bucket

        self.resource_dir = resource_dir
        self.source_dir = source_dir
        self.priority = priority
        self.dir = None
        self.run_entry = None
        self.config = None
        self.info = None
        self.cout = ""
        self.cout_write_cursor = 0
        self.saved_metrics = {}

        from google.cloud import storage
        import google.auth.exceptions

        try:
            client = storage.Client()
        except google.auth.exceptions.DefaultCredentialsError:
            raise ConnectionError(
                "Could not create Google Cloud Storage observer, are you "
                "sure that you have set environment variable GOOGLE_APPLICATION_CREDENTIALS?"
            )

        self.bucket = client.bucket(bucket)

    def _objects_exist_in_dir(self, prefix):
        # This should be run after you've confirmed the bucket
        # exists, and will error out if it does not exist
        all_blobs = [blob for blob in self.bucket.list_blobs(prefix=prefix)]
        return len(all_blobs) > 0

    def _list_gcs_subdirs(self, prefix=None):
        if prefix is None:
            prefix = self.basedir

        iterator = self.bucket.list_blobs(prefix=prefix, delimiter="/")
        prefixes = set()
        for page in iterator.pages:
            prefixes.update(page.prefixes)

        return list(prefixes)

    def _determine_run_dir(self, _id):
        if _id is None:
            basepath = os.path.join(self.basedir, "")
            bucket_path_subdirs = self._list_gcs_subdirs(prefix=basepath)

            if not bucket_path_subdirs:
                max_run_id = 0
            else:
                relative_paths = [
                    path.replace(self.basedir, "").strip("/")
                    for path in bucket_path_subdirs
                ]
                integer_directories = [int(d) for d in relative_paths if d.isdigit()]
                if not integer_directories:
                    max_run_id = 0
                else:
                    # If there are directories under basedir that aren't
                    # numeric run directories, ignore those
                    max_run_id = max(integer_directories)

            _id = max_run_id + 1

        self.dir = gcs_join(self.basedir, str(_id))
        if self._objects_exist_in_dir(self.dir):
            raise FileExistsError("GCS dir at {} already exists".format(self.dir))
        return _id

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        _id = self._determine_run_dir(_id)

        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "meta": meta_info,
            "status": "QUEUED",
        }
        self.config = config
        self.info = {}

        self.save_json(self.run_entry, "run.json")
        self.save_json(self.config, "config.json")

        for s, m in ex_info["sources"]:
            self.save_file(s)

        return _id

    def save_sources(self, ex_info):
        base_dir = ex_info["base_dir"]
        source_info = []
        for s, m in ex_info["sources"]:
            abspath = os.path.join(base_dir, s)
            store_path, md5sum = self.find_or_save(abspath, self.source_dir)
            source_info.append([s, os.path.relpath(store_path, self.basedir)])
        return source_info

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):

        _id = self._determine_run_dir(_id)

        ex_info["sources"] = self.save_sources(ex_info)

        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "start_time": start_time.isoformat(),
            "meta": meta_info,
            "status": "RUNNING",
            "resources": [],
            "artifacts": [],
            "heartbeat": None,
        }
        self.config = config
        self.info = {}
        self.cout = ""
        self.cout_write_cursor = 0

        self.save_json(self.run_entry, "run.json")
        self.save_json(self.config, "config.json")
        self.save_cout()

        return _id

    def find_or_save(self, filename, store_dir):
        source_name, ext = os.path.splitext(os.path.basename(filename))
        md5sum = get_digest(filename)
        store_name = source_name + "_" + md5sum + ext
        store_path = gcs_join(store_dir, store_name)
        if len(self._list_gcs_subdirs(prefix=store_path)) == 0:
            self.save_file_to_base(filename, store_path)
        return store_path, md5sum

    def put_data(self, key, binary_data):
        blob = self.bucket.blob(key)
        blob.upload_from_file(binary_data)

    def save_json(self, obj, filename):
        key = gcs_join(self.dir, filename)
        blob = self.bucket.blob(key)
        blob.upload_from_string(
            json.dumps(flatten(obj), sort_keys=True, indent=2), content_type="text/json"
        )

    def save_file(self, filename, target_name=None):
        target_name = target_name or os.path.basename(filename)
        key = gcs_join(self.dir, target_name)
        self.put_data(key, open(filename, "rb"))

    def save_file_to_base(self, filename, target_name=None):
        target_name = target_name or os.path.basename(filename)
        self.put_data(target_name, open(filename, "rb"))

    def save_directory(self, source_dir, target_name):
        target_name = target_name or os.path.basename(source_dir)
        all_files = []
        for root, dirs, files in os.walk(source_dir):
            all_files += [os.path.join(root, f) for f in files]

        for filename in all_files:
            file_location = gcs_join(
                self.dir, target_name, os.path.relpath(filename, source_dir)
            )
            self.put_data(file_location, open(filename, "rb"))

    def save_cout(self):
        binary_data = self.cout[self.cout_write_cursor :].encode("utf-8")
        key = gcs_join(self.dir, "cout.txt")
        blob = self.bucket.blob(key)
        blob.upload_from_string(binary_data, content_type="text/plain")
        self.cout_write_cursor = len(self.cout)

    def heartbeat_event(self, info, captured_out, beat_time, result):
        self.info = info
        self.run_entry["heartbeat"] = beat_time.isoformat()
        self.run_entry["result"] = result
        self.cout = captured_out
        self.save_cout()
        self.save_json(self.run_entry, "run.json")
        if self.info:
            self.save_json(self.info, "info.json")

    def completed_event(self, stop_time, result):
        self.run_entry["stop_time"] = stop_time.isoformat()
        self.run_entry["result"] = result
        self.run_entry["status"] = "COMPLETED"

        self.save_json(self.run_entry, "run.json")

    def interrupted_event(self, interrupt_time, status):
        self.run_entry["stop_time"] = interrupt_time.isoformat()
        self.run_entry["status"] = status
        self.save_json(self.run_entry, "run.json")

    def failed_event(self, fail_time, fail_trace):
        self.run_entry["stop_time"] = fail_time.isoformat()
        self.run_entry["status"] = "FAILED"
        self.run_entry["fail_trace"] = fail_trace
        self.save_json(self.run_entry, "run.json")

    def resource_event(self, filename):
        store_path, md5sum = self.find_or_save(filename, self.resource_dir)
        self.run_entry["resources"].append([filename, store_path])
        self.save_json(self.run_entry, "run.json")

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        self.save_file(filename, name)
        self.run_entry["artifacts"].append(name)
        self.save_json(self.run_entry, "run.json")

    def log_metrics(self, metrics_by_name, info):
        """Store new measurements into metrics.json."""
        for metric_name, metric_ptr in metrics_by_name.items():

            if metric_name not in self.saved_metrics:
                self.saved_metrics[metric_name] = {
                    "values": [],
                    "steps": [],
                    "timestamps": [],
                }

            self.saved_metrics[metric_name]["values"] += metric_ptr["values"]
            self.saved_metrics[metric_name]["steps"] += metric_ptr["steps"]

            timestamps_norm = [ts.isoformat() for ts in metric_ptr["timestamps"]]
            self.saved_metrics[metric_name]["timestamps"] += timestamps_norm

        self.save_json(self.saved_metrics, "metrics.json")

    def __eq__(self, other):
        if isinstance(other, GoogleCloudStorageObserver):
            return self.bucket_id == other.bucket_id and self.basedir == other.basedir
        else:
            return False


@cli_option("-G", "--gcs")
def gcs_option(args, run):
    """Add a Google Cloud Storage File observer to the experiment.

    The argument value should be `gs://<bucket>/path/to/exp`.
    """
    match_obj = re.match(r"gs:\/\/([^\/]*)\/(.*)", args)
    if match_obj is None or len(match_obj.groups()) != 2:
        raise ValueError(
            "Valid bucket specification not found. "
            "Enter bucket and directory path like: "
            "gs://<bucket>/path/to/exp"
        )
    bucket, basedir = match_obj.groups()
    run.observers.append(GoogleCloudStorageObserver(bucket=bucket, basedir=basedir))



#!/usr/bin/env python
# coding=utf-8

import random

import sacred.optional as opt
from sacred.utils import module_is_in_cache

SEEDRANGE = (1, int(1e9))


def get_seed(rnd=None):
    if rnd is None:
        return random.randint(*SEEDRANGE)
    return rnd.randint(*SEEDRANGE)


def create_rnd(seed):
    assert isinstance(seed, int), "Seed has to be integer but was {} {}".format(
        repr(seed), type(seed)
    )
    if opt.has_numpy:
        return opt.np.random.RandomState(seed)
    else:
        return random.Random(seed)


def set_global_seed(seed):
    random.seed(seed)
    if opt.has_numpy:
        opt.np.random.seed(seed)
    if module_is_in_cache("tensorflow"):
        tf = opt.get_tensorflow()
        tf.set_random_seed(seed)
    if module_is_in_cache("torch"):
        import torch

        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)



#!/usr/bin/env python
# coding=utf-8

import collections
import contextlib
import importlib
import logging
import pkgutil
import re
import shlex
import sys
import threading
import traceback as tb
from functools import partial
from packaging import version
from typing import Union
from pathlib import Path

import wrapt

__all__ = [
    "NO_LOGGER",
    "PYTHON_IDENTIFIER",
    "CircularDependencyError",
    "ObserverError",
    "SacredInterrupt",
    "TimeoutInterrupt",
    "create_basic_stream_logger",
    "recursive_update",
    "iterate_flattened",
    "iterate_flattened_separately",
    "set_by_dotted_path",
    "get_by_dotted_path",
    "iter_prefixes",
    "join_paths",
    "is_prefix",
    "convert_to_nested_dict",
    "convert_camel_case_to_snake_case",
    "print_filtered_stacktrace",
    "optional_kwargs_decorator",
    "get_inheritors",
    "apply_backspaces_and_linefeeds",
    "rel_path",
    "IntervalTimer",
    "PathType",
]

NO_LOGGER = logging.getLogger("ignore")
NO_LOGGER.disabled = 1

PATHCHANGE = object()

PYTHON_IDENTIFIER = re.compile("^[a-zA-Z_][_a-zA-Z0-9]*$")

PathType = Union[str, bytes, Path]


class ObserverError(Exception):
    """Error that an observer raises but that should not make the run fail."""


class SacredInterrupt(Exception):
    """Base-Class for all custom interrupts.

    For more information see :ref:`custom_interrupts`.
    """

    STATUS = "INTERRUPTED"


class TimeoutInterrupt(SacredInterrupt):
    """Signal that the experiment timed out.

    This exception can be used in client code to indicate that the run
    exceeded its time limit and has been interrupted because of that.
    The status of the interrupted run will then be set to ``TIMEOUT``.

    For more information see :ref:`custom_interrupts`.
    """

    STATUS = "TIMEOUT"


class SacredError(Exception):
    def __init__(
        self,
        message,
        print_traceback=True,
        filter_traceback="default",
        print_usage=False,
    ):
        super().__init__(message)
        self.print_traceback = print_traceback
        if filter_traceback not in ["always", "default", "never"]:
            raise ValueError(
                "filter_traceback must be one of 'always', "
                "'default' or 'never', not " + filter_traceback
            )
        self.filter_traceback = filter_traceback
        self.print_usage = print_usage


class CircularDependencyError(SacredError):
    """The ingredients of the current experiment form a circular dependency."""

    @classmethod
    @contextlib.contextmanager
    def track(cls, ingredient):
        try:
            yield
        except CircularDependencyError as e:
            if not e.__circular_dependency_handled__:
                if ingredient in e.__ingredients__:
                    e.__circular_dependency_handled__ = True
                e.__ingredients__.append(ingredient)
            raise e

    def __init__(
        self,
        message="Circular dependency detected:",
        ingredients=None,
        print_traceback=True,
        filter_traceback="default",
        print_usage=False,
    ):
        super().__init__(
            message,
            print_traceback=print_traceback,
            filter_traceback=filter_traceback,
            print_usage=print_usage,
        )

        if ingredients is None:
            ingredients = []
        self.__ingredients__ = ingredients
        self.__circular_dependency_handled__ = False

    def __str__(self):
        return super().__str__() + "->".join(
            [i.path for i in reversed(self.__ingredients__)]
        )


class ConfigError(SacredError):
    """Pretty prints the conflicting configuration values."""

    def __init__(
        self,
        message,
        conflicting_configs=(),
        print_conflicting_configs=True,
        print_traceback=True,
        filter_traceback="default",
        print_usage=False,
        config=None,
    ):
        super().__init__(
            message,
            print_traceback=print_traceback,
            filter_traceback=filter_traceback,
            print_usage=print_usage,
        )
        self.print_conflicting_configs = print_conflicting_configs

        if isinstance(conflicting_configs, str):
            conflicting_configs = (conflicting_configs,)

        self.__conflicting_configs__ = conflicting_configs
        self.__prefix_handled__ = False

        if config is None:
            config = {}
        self.__config__ = config

    @classmethod
    @contextlib.contextmanager
    def track(cls, wrapped):
        try:
            yield
        except ConfigError as e:
            if not e.__prefix_handled__:
                if wrapped.prefix:
                    e.__conflicting_configs__ = (
                        join_paths(wrapped.prefix, str(c))
                        for c in e.__conflicting_configs__
                    )
                e.__config__ = wrapped.config
                e.__prefix_handled__ = True
            raise e

    def __str__(self):
        s = super().__str__()
        if self.print_conflicting_configs:
            # Add a list formatted as below to the string s:
            #
            # Conflicting configuration values:
            #   a=3
            #   b.c=4
            s += "\nConflicting configuration values:"
            for conflicting_config in self.__conflicting_configs__:
                s += "\n  {}={}".format(
                    conflicting_config,
                    get_by_dotted_path(self.__config__, conflicting_config),
                )
        return s


class InvalidConfigError(ConfigError):
    """Can be raised in the user code if an error in the configuration is detected.

    Examples
    --------
    >>> # Experiment definitions ...
    ... @ex.automain
    ... def main(a, b):
    ...     if a != b['a']:
    ...         raise InvalidConfigError(
    ...                     'Need to be equal',
    ...                     conflicting_configs=('a', 'b.a'))
    """

    pass


class MissingConfigError(SacredError):
    """A config value that is needed by a captured function is not present in the provided config."""

    def __init__(
        self,
        message="Configuration values are missing:",
        missing_configs=(),
        print_traceback=False,
        filter_traceback="default",
        print_usage=True,
    ):
        message = "{} {}".format(message, missing_configs)
        super().__init__(
            message,
            print_traceback=print_traceback,
            filter_traceback=filter_traceback,
            print_usage=print_usage,
        )


class NamedConfigNotFoundError(SacredError):
    """A named config is not found."""

    def __init__(
        self,
        named_config,
        message="Named config not found:",
        available_named_configs=(),
        print_traceback=False,
        filter_traceback="default",
        print_usage=False,
    ):
        message = '{} "{}". Available config values are: {}'.format(
            message, named_config, available_named_configs
        )
        super().__init__(
            message,
            print_traceback=print_traceback,
            filter_traceback=filter_traceback,
            print_usage=print_usage,
        )


class ConfigAddedError(ConfigError):
    SPECIAL_ARGS = {"_log", "_config", "_seed", "__doc__", "config_filename", "_run"}
    """Special args that show up in the captured args but can never be set
    by the user"""

    def __init__(
        self,
        conflicting_configs,
        message="Added new config entry that is not used anywhere",
        captured_args=(),
        print_conflicting_configs=True,
        print_traceback=False,
        filter_traceback="default",
        print_usage=False,
        print_suggestions=True,
        config=None,
    ):
        super().__init__(
            message,
            conflicting_configs=conflicting_configs,
            print_conflicting_configs=print_conflicting_configs,
            print_traceback=print_traceback,
            filter_traceback=filter_traceback,
            print_usage=print_usage,
            config=config,
        )
        self.captured_args = captured_args
        self.print_suggestions = print_suggestions

    def __str__(self):
        s = super().__str__()
        if self.print_suggestions:
            possible_keys = set(self.captured_args) - self.SPECIAL_ARGS
            if possible_keys:
                s += "\nPossible config keys are: {}".format(possible_keys)
        return s


class SignatureError(SacredError, TypeError):
    """Error that is raised when the passed arguments do not match the functions signature."""

    def __init__(
        self,
        message,
        print_traceback=True,
        filter_traceback="always",
        print_usage=False,
    ):
        super().__init__(message, print_traceback, filter_traceback, print_usage)


def create_basic_stream_logger():
    """Sets up a basic stream logger.

    Configures the root logger to use a
    `logging.StreamHandler` and sets the logging level to `logging.INFO`.

    Notes
    -----
        This does not change the logger configuration if the root logger
        already is configured (i.e. `len(getLogger().handlers) > 0`)
    """
    logging.basicConfig(
        level=logging.INFO, format="%(levelname)s - %(name)s - %(message)s"
    )
    return logging.getLogger("")


def recursive_update(d, u):
    """
    Given two dictionaries d and u, update dict d recursively.

    E.g.:
    d = {'a': {'b' : 1}}
    u = {'c': 2, 'a': {'d': 3}}
    => {'a': {'b': 1, 'd': 3}, 'c': 2}
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = recursive_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def iterate_flattened_separately(dictionary, manually_sorted_keys=None):
    """
    Recursively iterate over the items of a dictionary in a special order.

    First iterate over manually sorted keys and then over all items that are
    non-dictionary values (sorted by keys), then over the rest
    (sorted by keys), providing full dotted paths for every leaf.
    """
    manually_sorted_keys = manually_sorted_keys or []

    def get_order(key_and_value):
        key, value = key_and_value
        if key in manually_sorted_keys:
            return 0, manually_sorted_keys.index(key)
        elif not is_non_empty_dict(value):
            return 1, key
        else:
            return 2, key

    for key, value in sorted(dictionary.items(), key=get_order):
        if is_non_empty_dict(value):
            yield key, PATHCHANGE
            for k, val in iterate_flattened_separately(value, manually_sorted_keys):
                yield join_paths(key, k), val
        else:
            yield key, value


def is_non_empty_dict(python_object):
    return isinstance(python_object, dict) and python_object


def iterate_flattened(d):
    """
    Recursively iterate over the items of a dictionary.

    Provides a full dotted paths for every leaf.
    """
    for key in sorted(d.keys()):
        value = d[key]
        if isinstance(value, dict) and value:
            for k, v in iterate_flattened(d[key]):
                yield join_paths(key, k), v
        else:
            yield key, value


def set_by_dotted_path(d, path, value):
    """
    Set an entry in a nested dict using a dotted path.

    Will create dictionaries as needed.

    Examples
    --------
    >>> d = {'foo': {'bar': 7}}
    >>> set_by_dotted_path(d, 'foo.bar', 10)
    >>> d
    {'foo': {'bar': 10}}
    >>> set_by_dotted_path(d, 'foo.d.baz', 3)
    >>> d
    {'foo': {'bar': 10, 'd': {'baz': 3}}}

    """
    split_path = path.split(".")
    current_option = d
    for p in split_path[:-1]:
        if p not in current_option:
            current_option[p] = dict()
        current_option = current_option[p]
    current_option[split_path[-1]] = value


def get_by_dotted_path(d, path, default=None):
    """
    Get an entry from nested dictionaries using a dotted path.

    Example
    -------
    >>> get_by_dotted_path({'foo': {'a': 12}}, 'foo.a')
    12
    """
    if not path:
        return d
    split_path = path.split(".")
    current_option = d
    for p in split_path:
        if p not in current_option:
            return default
        current_option = current_option[p]
    return current_option


def iter_prefixes(path):
    """
    Iterate through all (non-empty) prefixes of a dotted path.

    Example
    -------
    >>> list(iter_prefixes('foo.bar.baz'))
    ['foo', 'foo.bar', 'foo.bar.baz']
    """
    split_path = path.split(".")
    for i in range(1, len(split_path) + 1):
        yield join_paths(*split_path[:i])


def join_paths(*parts):
    """Join different parts together to a valid dotted path."""
    return ".".join(str(p).strip(".") for p in parts if p)


def is_prefix(pre_path, path):
    """Return True if pre_path is a path-prefix of path."""
    pre_path = pre_path.strip(".")
    path = path.strip(".")
    return not pre_path or path.startswith(pre_path + ".")


def rel_path(base, path):
    """Return path relative to base."""
    if base == path:
        return ""
    assert is_prefix(base, path), "{} not a prefix of {}".format(base, path)
    return path[len(base) :].strip(".")


def convert_to_nested_dict(dotted_dict):
    """Convert a dict with dotted path keys to corresponding nested dict."""
    nested_dict = {}
    for k, v in iterate_flattened(dotted_dict):
        set_by_dotted_path(nested_dict, k, v)
    return nested_dict


def _is_sacred_frame(frame):
    return frame.f_globals["__name__"].split(".")[0] == "sacred"


def print_filtered_stacktrace(filter_traceback="default"):
    print(format_filtered_stacktrace(filter_traceback), file=sys.stderr)


def format_filtered_stacktrace(filter_traceback="default"):
    """
    Returns the traceback as `string`.

    `filter_traceback` can be one of:
        - 'always': always filter out sacred internals
        - 'default': Default behaviour: filter out sacred internals
                if the exception did not originate from within sacred, and
                print just the internal stack trace otherwise
        - 'never': don't filter, always print full traceback
        - All other values will fall back to 'never'.
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    # determine if last exception is from sacred
    current_tb = exc_traceback
    while current_tb.tb_next is not None:
        current_tb = current_tb.tb_next

    if filter_traceback == "default" and _is_sacred_frame(current_tb.tb_frame):
        # just print sacred internal trace
        header = [
            "Exception originated from within Sacred.\n"
            "Traceback (most recent calls):\n"
        ]
        texts = tb.format_exception(exc_type, exc_value, current_tb)
        return "".join(header + texts[1:]).strip()
    elif filter_traceback in ("default", "always"):
        # print filtered stacktrace
        if sys.version_info >= (3, 5):
            tb_exception = tb.TracebackException(
                exc_type, exc_value, exc_traceback, limit=None
            )
            return "".join(filtered_traceback_format(tb_exception))
        else:
            s = "Traceback (most recent calls WITHOUT Sacred internals):"
            current_tb = exc_traceback
            while current_tb is not None:
                if not _is_sacred_frame(current_tb.tb_frame):
                    tb.print_tb(current_tb, 1)
                current_tb = current_tb.tb_next
            s += "\n".join(tb.format_exception_only(exc_type, exc_value)).strip()
            return s
    elif filter_traceback == "never":
        # print full stacktrace
        return "\n".join(tb.format_exception(exc_type, exc_value, exc_traceback))
    else:
        raise ValueError("Unknown value for filter_traceback: " + filter_traceback)


def format_sacred_error(e, short_usage):
    lines = []
    if e.print_usage:
        lines.append(short_usage)
    if e.print_traceback:
        lines.append(format_filtered_stacktrace(e.filter_traceback))
    else:
        lines.append("\n".join(tb.format_exception_only(type(e), e)))
    return "\n".join(lines)


def filtered_traceback_format(tb_exception, chain=True):
    if chain:
        if tb_exception.__cause__ is not None:
            yield from filtered_traceback_format(tb_exception.__cause__, chain=chain)
            yield tb._cause_message
        elif (
            tb_exception.__context__ is not None
            and not tb_exception.__suppress_context__
        ):
            yield from filtered_traceback_format(tb_exception.__context__, chain=chain)
            yield tb._context_message
    yield "Traceback (most recent calls WITHOUT Sacred internals):\n"
    current_tb = tb_exception.exc_traceback
    while current_tb is not None:
        if not _is_sacred_frame(current_tb.tb_frame):
            stack = tb.StackSummary.extract(
                tb.walk_tb(current_tb), limit=1, lookup_lines=True, capture_locals=False
            )
            yield from stack.format()
        current_tb = current_tb.tb_next
    yield from tb_exception.format_exception_only()


# noinspection PyUnusedLocal
@wrapt.decorator
def optional_kwargs_decorator(wrapped, instance=None, args=None, kwargs=None):
    # here wrapped is itself a decorator
    if args:  # means it was used as a normal decorator (so just call it)
        return wrapped(*args, **kwargs)
    else:  # used with kwargs, so we need to return a decorator
        return partial(wrapped, **kwargs)


def get_inheritors(cls):
    """Get a set of all classes that inherit from the given class."""
    subclasses = set()
    work = [cls]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


# Credit to Zarathustra and epost from stackoverflow
# Taken from http://stackoverflow.com/a/1176023/1388435
def convert_camel_case_to_snake_case(name):
    """Convert CamelCase to snake_case."""
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def apply_backspaces_and_linefeeds(text):
    """
    Interpret backspaces and linefeeds in text like a terminal would.

    Interpret text like a terminal by removing backspace and linefeed
    characters and applying them line by line.

    If final line ends with a carriage it keeps it to be concatenable with next
    output chunk.
    """
    orig_lines = text.split("\n")
    orig_lines_len = len(orig_lines)
    new_lines = []
    for orig_line_idx, orig_line in enumerate(orig_lines):
        chars, cursor = [], 0
        orig_line_len = len(orig_line)
        for orig_char_idx, orig_char in enumerate(orig_line):
            if orig_char == "\r" and (
                orig_char_idx != orig_line_len - 1
                or orig_line_idx != orig_lines_len - 1
            ):
                cursor = 0
            elif orig_char == "\b":
                cursor = max(0, cursor - 1)
            else:
                if (
                    orig_char == "\r"
                    and orig_char_idx == orig_line_len - 1
                    and orig_line_idx == orig_lines_len - 1
                ):
                    cursor = len(chars)
                if cursor == len(chars):
                    chars.append(orig_char)
                else:
                    chars[cursor] = orig_char
                cursor += 1
        new_lines.append("".join(chars))
    return "\n".join(new_lines)


def module_exists(modname):
    """Checks if a module exists without actually importing it."""
    try:
        return pkgutil.find_loader(modname) is not None
    except ImportError:
        # TODO: Temporary fix for tf 1.14.0.
        # Should be removed once fixed in tf.
        return True


def modules_exist(*modnames):
    return all(module_exists(m) for m in modnames)


def module_is_in_cache(modname):
    """Checks if a module was imported before (is in the import cache)."""
    return modname in sys.modules


def parse_version(version_string):
    """Returns a parsed version string."""
    return version.parse(version_string)


def get_package_version(name):
    """Returns a parsed version string of a package."""
    version_string = importlib.import_module(name).__version__
    return parse_version(version_string)


def ensure_wellformed_argv(argv):
    if argv is None:
        argv = sys.argv
    elif isinstance(argv, str):
        argv = shlex.split(argv)
    else:
        if not isinstance(argv, (list, tuple)):
            raise ValueError("argv must be str or list, but was {}".format(type(argv)))
        if not all([isinstance(a, str) for a in argv]):
            problems = [a for a in argv if not isinstance(a, str)]
            raise ValueError(
                "argv must be list of str but contained the "
                "following elements: {}".format(problems)
            )
    return argv


class IntervalTimer(threading.Thread):
    @classmethod
    def create(cls, func, interval=10):
        stop_event = threading.Event()
        timer_thread = cls(stop_event, func, interval)
        return stop_event, timer_thread

    def __init__(self, event, func, interval=10.0):
        super().__init__()
        self.stopped = event
        self.func = func
        self.interval = interval

    def run(self):
        while not self.stopped.wait(self.interval):
            self.func()
        self.func()



#!/usr/bin/env python
# coding=utf-8

import os
import sys
import subprocess
from io import StringIO
from contextlib import contextmanager
import wrapt
from sacred.optional import libc
from tempfile import NamedTemporaryFile
from sacred.settings import SETTINGS


def flush():
    """Try to flush all stdio buffers, both from python and from C."""
    try:
        sys.stdout.flush()
        sys.stderr.flush()
    except (AttributeError, ValueError, OSError):
        pass  # unsupported
    try:
        libc.fflush(None)
    except (AttributeError, ValueError, OSError):
        pass  # unsupported


def get_stdcapturer(mode=None):
    mode = mode if mode is not None else SETTINGS.CAPTURE_MODE
    capture_options = {"no": no_tee, "fd": tee_output_fd, "sys": tee_output_python}
    if mode not in capture_options:
        raise KeyError(
            "Unknown capture mode '{}'. Available options are {}".format(
                mode, sorted(capture_options.keys())
            )
        )
    return mode, capture_options[mode]


class TeeingStreamProxy(wrapt.ObjectProxy):
    """A wrapper around stdout or stderr that duplicates all output to out."""

    def __init__(self, wrapped, out):
        super().__init__(wrapped)
        self._self_out = out

    def write(self, data):
        self.__wrapped__.write(data)
        self._self_out.write(data)

    def flush(self):
        self.__wrapped__.flush()
        self._self_out.flush()


class CapturedStdout:
    def __init__(self, buffer):
        self.buffer = buffer
        self.read_position = 0
        self.final = None

    @property
    def closed(self):
        return self.buffer.closed

    def flush(self):
        return self.buffer.flush()

    def get(self):
        if self.final is None:
            self.buffer.seek(self.read_position)
            value = self.buffer.read()
            self.read_position = self.buffer.tell()
            return value
        else:
            value = self.final
            self.final = None
            return value

    def finalize(self):
        self.flush()
        self.final = self.get()
        self.buffer.close()


@contextmanager
def no_tee():
    out = CapturedStdout(StringIO())
    try:
        yield out
    finally:
        out.finalize()


@contextmanager
def tee_output_python():
    """Duplicate sys.stdout and sys.stderr to new StringIO."""
    buffer = StringIO()
    out = CapturedStdout(buffer)
    orig_stdout, orig_stderr = sys.stdout, sys.stderr
    flush()
    sys.stdout = TeeingStreamProxy(sys.stdout, buffer)
    sys.stderr = TeeingStreamProxy(sys.stderr, buffer)
    try:
        yield out
    finally:
        flush()
        out.finalize()
        sys.stdout, sys.stderr = orig_stdout, orig_stderr


# Duplicate stdout and stderr to a file. Inspired by:
# http://eli.thegreenplace.net/2015/redirecting-all-kinds-of-stdout-in-python/
# http://stackoverflow.com/a/651718/1388435
# http://stackoverflow.com/a/22434262/1388435
@contextmanager
def tee_output_fd():
    """Duplicate stdout and stderr to a file on the file descriptor level."""
    with NamedTemporaryFile(mode="w+") as target:
        original_stdout_fd = 1
        original_stderr_fd = 2
        target_fd = target.fileno()

        # Save a copy of the original stdout and stderr file descriptors
        saved_stdout_fd = os.dup(original_stdout_fd)
        saved_stderr_fd = os.dup(original_stderr_fd)

        try:
            # start_new_session=True to move process to a new process group
            # this is done to avoid receiving KeyboardInterrupts (see #149)
            tee_stdout = subprocess.Popen(
                ["tee", "-a", target.name],
                start_new_session=True,
                stdin=subprocess.PIPE,
                stdout=1,
            )
            tee_stderr = subprocess.Popen(
                ["tee", "-a", target.name],
                start_new_session=True,
                stdin=subprocess.PIPE,
                stdout=2,
            )
        except (FileNotFoundError, OSError, AttributeError):
            # No tee found in this operating system. Trying to use a python
            # implementation of tee. However this is slow and error-prone.
            tee_stdout = subprocess.Popen(
                [sys.executable, "-m", "sacred.pytee"],
                stdin=subprocess.PIPE,
                stderr=target_fd,
            )
            tee_stderr = subprocess.Popen(
                [sys.executable, "-m", "sacred.pytee"],
                stdin=subprocess.PIPE,
                stdout=target_fd,
            )

        flush()
        os.dup2(tee_stdout.stdin.fileno(), original_stdout_fd)
        os.dup2(tee_stderr.stdin.fileno(), original_stderr_fd)
        out = CapturedStdout(target)

        try:
            yield out  # let the caller do their printing
        finally:
            flush()

            # then redirect stdout back to the saved fd
            tee_stdout.stdin.close()
            tee_stderr.stdin.close()

            # restore original fds
            os.dup2(saved_stdout_fd, original_stdout_fd)
            os.dup2(saved_stderr_fd, original_stderr_fd)

            tee_stdout.wait(timeout=1)
            tee_stderr.wait(timeout=1)

            os.close(saved_stdout_fd)
            os.close(saved_stderr_fd)
            out.finalize()
